from typing import Callable, Optional

from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.consumer.reply import Reply

MessageHandler = Callable[[PikaMessage], Optional[Reply]]



from typing import List, Type


class Doc:
    def __init__(self) -> None:
        self._commands: List[str] = []
        self._command_replies: List[str] = []
        self._events: List[str] = []

    def add_event(self, event_source: str, event_type: Type) -> None:
        self._events.append(f"{event_source}: {event_type.__name__}")

    def add_command(self, command_type: Type) -> None:
        self._commands.append(command_type.__name__)

    def add_command_reply(self, command_destination: str, command_type: Type) -> None:
        self._command_replies.append(
            f"{command_type.__name__} from {command_destination}"
        )

    def get_commands(self) -> List[str]:
        return self._commands

    def get_events(self) -> List[str]:
        return self._events

    def get_command_replies(self) -> List[str]:
        return self._command_replies



from collections import defaultdict
from contextlib import ExitStack, contextmanager
from typing import Callable, ContextManager, DefaultDict, List, Optional, Union

from myrabbit.core.consumer.pika_message import PikaMessage

Callback = Callable
Middleware = ContextManager


class Callbacks:
    def __init__(self, callbacks: Optional[DefaultDict[str, List[Callback]]] = None):
        self._callbacks = callbacks or defaultdict(list)

    def add_callback(self, name: str, callback: Union[Callback, Middleware]) -> None:
        self._callbacks[name].append(callback)

    def before_request(self, message: PikaMessage) -> None:
        for cb in self._callbacks.get("before_request", []):
            cb(self, message)

    def after_request(self, message: PikaMessage) -> None:
        for cb in self._callbacks.get("after_request", []):
            cb(self, message)

    def run_middleware(self, message: PikaMessage) -> ExitStack:
        with ExitStack() as stack:
            for middleware in self._callbacks.get("middleware", []):
                cm = contextmanager(middleware)
                stack.enter_context(cm(self, message))

            return stack.pop_all()

    def __eq__(self, other):
        if not isinstance(other, Callbacks):
            return False
        return self._callbacks == other._callbacks



from functools import wraps
from typing import Callable, List, Optional, Type

from pika import BasicProperties

from myrabbit.core.consumer.listener import Listener
from myrabbit.core.converter import DEFAULT_CONVERTERS, Converter
from myrabbit.events.event_bus import EventBus
from myrabbit.events.event_with_message import EventType, EventWithMessage
from myrabbit.events.listen_event_strategy import ListenEventStrategy


class EventBusAdapter:
    def __init__(
        self, event_bus: EventBus, converters: Optional[List[Converter]] = None,
    ):
        self.event_bus = event_bus
        self.converters = converters or DEFAULT_CONVERTERS

    def get_converter(self, event: EventType) -> Converter:
        for converter in self.converters:
            if converter.accepts(event):
                return converter
        raise RuntimeError("Converter not found")

    def publish(
        self,
        event_source: str,
        event: EventType,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        event_name, body = self.get_converter(event).name_and_body(event)
        self.event_bus.publish(event_source, event_name, body, properties)

    def listener(
        self,
        event_destination: str,
        event_source: str,
        event_type: Type[EventType],
        callback: Callable[[EventWithMessage], None],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
        listen_strategy: Optional[ListenEventStrategy] = None,
        method_name: Optional[str] = None,
    ) -> Listener:
        converter = self.get_converter(event_type)

        @wraps(callback)
        def instantiate_event(event_with_message: EventWithMessage) -> None:
            event_with_message.event = converter.instantiate(
                event_type, event_with_message.event
            )
            callback(event_with_message)

        event_name = converter.name(event_type)

        return self.event_bus.listener(
            event_destination=event_destination,
            event_source=event_source,
            event_name=event_name,
            callback=instantiate_event,
            exchange_params=exchange_params,
            queue_params=queue_params,
            listen_strategy=listen_strategy,
            method_name=method_name,
        )
